// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file ITL_OEMSolver.h */


#ifndef AMDIS_ITL_OEM_SOLVER_H
#define AMDIS_ITL_OEM_SOLVER_H

#include <iostream>
#include "OEMSolver.h"
#include "ITL_Preconditioner.h"
#include "SolverMatrix.h"
#include "DOFVector.h"
#include "SystemVector.h"
#include "DOFMatrix.h"
#include "Parameters.h"

#include <boost/numeric/itl/itl.hpp>
#include <boost/numeric/mtl/mtl.hpp>

namespace AMDiS {

  
  /**
   * \ingroup Solver
   * 
   *\brief
   * Template for using solvers from ITL in AMDiS. 
   */
  template <typename ITLSolver>
  class ITL_OEMSolver : public OEMSolver
  {
    typedef ITL_Preconditioner<itl::pc::identity<DOFMatrix> > id_t;

  public:
    /// The constructor reads needed parameters and sets solvers \ref name.
    ITL_OEMSolver(std::string name) : OEMSolver(name)  {}
  
    ~ITL_OEMSolver() {}


    /// Solves the system iteratively
    // This function is virtual because derived class calls solver with parameter
    virtual int solveSystem(const DOFMatrix::base_matrix_type&      A,
			    mtl::dense_vector<value_type>&          x,
			    const mtl::dense_vector<value_type>&    b)
    {
      itl::cyclic_iteration<value_type> iter(b, this->max_iter, this->relative, 
					     this->tolerance, this->print_cycle);

      error = ITLSolver()(A, x, b, *this->leftPrecon, *this->rightPrecon, iter);
      iterations = iter.iterations();
      residual = iter.resid();
      return error;
    }


    /// Creator class used in the OEMSolverMap.
    class Creator : public OEMSolverCreator
    {
    public:
      virtual ~Creator() {}
      
      /// Returns a new CGSolver object.
      OEMSolver* create() 
      { 
	return new ITL_OEMSolver<ITLSolver>(this->name); 
      }
    };
  
  };


  /**
   * \ingroup Solver
   * 
   *\brief
   * Template for using solvers from ITL that have one run-time parameter in AMDiS. 
   */
  template <typename ITLSolver>
  class ITL_OEMSolver_para
      : public OEMSolver // ITL_OEMSolver<ITLSolver>
  {
      typedef DOFMatrix::value_type value_type;
  public:
      /// The constructor reads needed parameters and sets solvers \ref name.
      ITL_OEMSolver_para(std::string name) 
	  : OEMSolver(name), ell(OEMSolver::max_iter) 
      {
	Parameters::get(name + "->ell", ell);
      }
      
      ~ITL_OEMSolver_para() {}

      /// Set parameter of iterative solver
    void setEll(int p) { ell= p; }

    /// Solves the system iteratively
    int solveSystem(const DOFMatrix::base_matrix_type&      A,
		    mtl::dense_vector<value_type>&          x,
		    const mtl::dense_vector<value_type>&    b)
    {
	itl::cyclic_iteration<value_type> iter(b, this->max_iter, this->relative, 
					       this->tolerance, this->print_cycle);
	error = ITLSolver()(A, x, b, *this->leftPrecon, *this->rightPrecon, iter, ell);
	iterations = iter.iterations();
	residual = iter.resid();
	return error;
    }

    class Creator : public OEMSolverCreator
    {
    public:
      virtual ~Creator() {}
      
      /// Returns a new CGSolver object.
      OEMSolver* create() 
      { 
	return new ITL_OEMSolver_para<ITLSolver>(this->name); 
      }
    }; 
      
  private:
    /// parameter for bicgstab_ell iterative solver    
    int ell;
  };


} // namespace AMDiS

#endif // AMDIS_ITL_OEM_SOLVER_H
